package ru.dbotthepony.mc.otm.data.world

import com.google.common.collect.ImmutableList
import com.mojang.serialization.Codec
import com.mojang.serialization.DataResult
import com.mojang.serialization.MapCodec
import com.mojang.serialization.codecs.RecordCodecBuilder
import net.minecraft.util.RandomSource
import net.minecraft.util.valueproviders.IntProvider
import ru.dbotthepony.mc.otm.core.random
import java.util.Optional

class OneOfProvider<T> private constructor(val children: List<T>, val selector: Optional<IntProvider> = Optional.empty()) {
	fun select(random: RandomSource): T {
		if (selector.isEmpty) {
			return children.random(random)
		}

		val index = selector.get().sample(random)

		if (index !in children.indices) {
			throw RuntimeException("Provided selector ($selector) provided index out of bounds: $index (bounds: ${children.indices})")
		}

		return children[index]
	}

	class Builder<T> {
		private var selector: IntProvider? = null
		private val inputs = ArrayList<T>()

		fun indexSelector(selector: IntProvider): Builder<T> {
			this.selector = selector
			return this
		}

		fun add(vararg provider: T): Builder<T> {
			inputs.addAll(provider)
			return this
		}

		fun copy(): Builder<T> {
			val new = Builder<T>()
			new.selector = selector
			new.inputs.addAll(inputs)
			return new
		}

		fun build(): OneOfProvider<T> {
			return OneOfProvider(ImmutableList.copyOf(inputs), Optional.ofNullable(selector))
		}
	}

	companion object {
		@JvmStatic
		fun <T> of(vararg providers: T): OneOfProvider<T> {
			return OneOfProvider(ImmutableList.copyOf(providers))
		}

		@JvmStatic
		fun <T> of(providers: Collection<T>): OneOfProvider<T> {
			return OneOfProvider(ImmutableList.copyOf(providers))
		}

		@JvmStatic
		fun <T> createCodec(childrenCodec: Codec<T>): MapCodec<OneOfProvider<T>> {
			return RecordCodecBuilder.mapCodec {
				it.group(
					Codec.list(childrenCodec, 1, Int.MAX_VALUE).fieldOf("children").forGetter<OneOfProvider<T>> { it.children },
					IntProvider.POSITIVE_CODEC.optionalFieldOf("selector").forGetter<OneOfProvider<T>> { it.selector },
				).apply(it, ::OneOfProvider)
			}.validate {
				if (it.selector.isEmpty || it.selector.get().minValue in it.children.indices && it.selector.get().maxValue in it.children.indices) {
					return@validate DataResult.success(it)
				} else {
					return@validate DataResult.error { "Provided number sampler ${it.selector.get()} can produce values outside of children list index range (list size: ${it.children.size})" }
				}
			}
		}

		fun <S, T> createCodec(childrenCodec: Codec<T>, to: (OneOfProvider<T>) -> S, from: (S) -> OneOfProvider<T>): MapCodec<S> {
			return createCodec(childrenCodec).xmap(to, from)
		}
	}
}
